@@ -1,30 +1,33 @@
 # coding: utf-8
 
 import calendar
+import json
 from datetime import datetime
 
-try:
-    import simplejson as json
-except ImportError:
-    import json
-
-try:  # celery 3.x
-    from celery.utils.timeutils import timezone
-except ImportError:  # celery 4.x
-    from celery.utils.time import timezone
-
 from celery.schedules import schedule, crontab
+from celery.utils.time import timezone, FixedOffset
+from dateutil.rrule import weekday
+
 from .schedules import rrule
 
 
 def to_timestamp(dt):
-    """ convert UTC datetime to seconds since the epoch """
-    return calendar.timegm(dt.timetuple())
+    """ convert local tz aware datetime to seconds since the epoch """
+    return calendar.timegm(dt.utctimetuple())
 
 
-def from_timestamp(seconds):
+def get_utcoffset_minutes(dt):
+    """ calculates timezone utc offset, returns minutes relative to utc """
+    utcoffset = dt.utcoffset()
+
+    # Python 3: utcoffset / timedelta(minutes=1)
+    return utcoffset.total_seconds() / 60 if utcoffset else 0
+
+
+def from_timestamp(seconds, tz_minutes=0):
     """ convert seconds since the epoch to an UTC aware datetime """
-    return datetime.fromtimestamp(seconds, tz=timezone.utc)
+    tz = FixedOffset(tz_minutes) if tz_minutes else timezone.utc
+    return datetime.fromtimestamp(seconds, tz=tz)
 
 
 class RedBeatJSONDecoder(json.JSONDecoder):
@@ -38,7 +41,12 @@ class RedBeatJSONDecoder(json.JSONDecoder):
         objtype = d.pop('__type__')
 
         if objtype == 'datetime':
-            return datetime(tzinfo=timezone.utc, **d)
+            zone = d.pop('timezone', 'UTC')
+            try:
+                tzinfo = FixedOffset(zone / 60)
+            except TypeError:
+                tzinfo = timezone.get_timezone(zone)
+            return datetime(tzinfo=tzinfo, **d)
 
         if objtype == 'interval':
             return schedule(run_every=d['every'], relative=d['relative'])
@@ -46,11 +54,16 @@ class RedBeatJSONDecoder(json.JSONDecoder):
         if objtype == 'crontab':
             return crontab(**d)
 
+        if objtype == 'weekday':
+            return weekday(**d)
+
         if objtype == 'rrule':
             # Decode timestamp values into datetime objects
-            for key in ['dtstart', 'until']:
-                if d[key] is not None:
-                    d[key] = from_timestamp(d[key])
+            for key, tz_key in [('dtstart', 'dtstart_tz'), ('until', 'until_tz')]:
+                timestamp = d.get(key)
+                tz_minutes = d.pop(tz_key, 0)
+                if timestamp is not None:
+                    d[key] = from_timestamp(timestamp, tz_minutes)
             return rrule(**d)
 
         d['__type__'] = objtype
@@ -61,6 +74,13 @@ class RedBeatJSONDecoder(json.JSONDecoder):
 class RedBeatJSONEncoder(json.JSONEncoder):
     def default(self, obj):
         if isinstance(obj, datetime):
+            if obj.tzinfo is None:
+                timezone = 'UTC'
+            elif obj.tzinfo.zone is None:
+                timezone = obj.tzinfo.utcoffset(None).total_seconds()
+            else:
+                timezone = obj.tzinfo.zone
+
             return {
                 '__type__': 'datetime',
                 'year': obj.year,
@@ -70,7 +90,9 @@ class RedBeatJSONEncoder(json.JSONEncoder):
                 'minute': obj.minute,
                 'second': obj.second,
                 'microsecond': obj.microsecond,
+                'timezone': timezone,
             }
+
         if isinstance(obj, crontab):
             return {
                 '__type__': 'crontab',
@@ -80,21 +102,14 @@ class RedBeatJSONEncoder(json.JSONEncoder):
                 'day_of_month': obj._orig_day_of_month,
                 'month_of_year': obj._orig_month_of_year,
             }
-        if isinstance(obj, rrule):
-            # Convert datetime objects to timestamps
-            dtstart_ts = to_timestamp(obj.dtstart) \
-                if obj.dtstart else None
-            until_ts = to_timestamp(obj.until) \
-                if obj.until else None
 
-            return {
+        if isinstance(obj, rrule):
+            res = {
                 '__type__': 'rrule',
                 'freq': obj.freq,
-                'dtstart': dtstart_ts,
                 'interval': obj.interval,
                 'wkst': obj.wkst,
                 'count': obj.count,
-                'until': until_ts,
                 'bysetpos': obj.bysetpos,
                 'bymonth': obj.bymonth,
                 'bymonthday': obj.bymonthday,
@@ -104,12 +119,28 @@ class RedBeatJSONEncoder(json.JSONEncoder):
                 'byweekday': obj.byweekday,
                 'byhour': obj.byhour,
                 'byminute': obj.byminute,
-                'bysecond': obj.bysecond
+                'bysecond': obj.bysecond,
             }
+
+            # Convert datetime objects to timestamps
+            if obj.dtstart:
+                res['dtstart'] = to_timestamp(obj.dtstart)
+                res['dtstart_tz'] = get_utcoffset_minutes(obj.dtstart)
+
+            if obj.until:
+                res['until'] = to_timestamp(obj.until)
+                res['until_tz'] = get_utcoffset_minutes(obj.until)
+
+            return res
+
+        if isinstance(obj, weekday):
+            return {'__type__': 'weekday', 'wkday': obj.weekday}
+
         if isinstance(obj, schedule):
             return {
                 '__type__': 'interval',
                 'every': obj.run_every.total_seconds(),
                 'relative': bool(obj.relative),
             }
+
         return super(RedBeatJSONEncoder, self).default(obj)
