@@ -5,63 +5,74 @@
 
 from __future__ import absolute_import
 
-import calendar
+import json
 import logging
 import warnings
+import ssl
 from datetime import datetime, MINYEAR
-from distutils.version import StrictVersion
 
-try:
-    import simplejson as json
-except ImportError:
-    import json
-
-from celery import VERSION as CELERY_VERSION
 from celery.beat import Scheduler, ScheduleEntry, DEFAULT_MAX_INTERVAL
 from celery.utils.log import get_logger
 from celery.signals import beat_init
-try:  # celery 3.x
-    from celery.utils.timeutils import humanize_seconds, timezone
-    from kombu.utils import cached_property
-except ImportError:  # celery 4.x
-    from celery.utils.time import humanize_seconds, timezone
-    from kombu.utils.objects import cached_property
+from celery.utils.time import humanize_seconds
+from kombu.utils.objects import cached_property
 from celery.app import app_or_default
-from celery.five import values
 from kombu.utils.url import maybe_sanitize_url
-from tenacity import (before_sleep_log,
-                      retry,
-                      retry_if_exception_type,
-                      stop_after_delay,
-                      wait_exponential)
-import pytz
+from tenacity import (
+    retry,
+    retry_if_exception_type,
+    stop_after_delay,
+    wait_exponential,
+)
 import redis.exceptions
 from redis.client import StrictRedis
-from redis import VERSION as REDIS_VERSION
 
-from .decoder import (
-    RedBeatJSONEncoder, RedBeatJSONDecoder,
-    from_timestamp, to_timestamp
-    )
+from .decoder import RedBeatJSONEncoder, RedBeatJSONDecoder, to_timestamp
 
-CELERY_4_OR_GREATER = CELERY_VERSION[0] >= 4
-REDIS_3_OR_GREATER = REDIS_VERSION[0] >= 3
+# Copied from:
+# https://github.com/andymccurdy/redis-py/blob/master/redis/lock.py#L33
+# Changes:
+#     The second line from the bottom: The original Lua script intends
+#     to extend time to (lock remaining time + additional time); while
+#     the script here extend time to a expected expiration time.
+# KEYS[1] - lock name
+# ARGS[1] - token
+# ARGS[2] - additional milliseconds
+# return 1 if the locks time was extended, otherwise 0
+LUA_EXTEND_TO_SCRIPT = """
+    local token = redis.call('get', KEYS[1])
+    if not token or token ~= ARGV[1] then
+        return 0
+    end
+    local expiration = redis.call('pttl', KEYS[1])
+    if not expiration then
+        expiration = 0
+    end
+    if expiration < 0 then
+        return 0
+    end
+    redis.call('pexpire', KEYS[1], ARGV[2])
+    return 1
+"""
 
 
 class RetryingConnection(object):
     """A proxy for the Redis connection that delegates all the calls to
     underlying Redis connection while retrying on connection or time-out error.
     """
+
     RETRY_MAX_WAIT = 30
 
     def __init__(self, retry_period, wrapped_connection):
         self.wrapped_connection = wrapped_connection
         self.retry_kwargs = dict(
-            retry=(retry_if_exception_type(redis.exceptions.ConnectionError)
-                   | retry_if_exception_type(redis.exceptions.TimeoutError)),
+            retry=(
+                retry_if_exception_type(redis.exceptions.ConnectionError)
+                | retry_if_exception_type(redis.exceptions.TimeoutError)
+            ),
             reraise=True,
             wait=wait_exponential(multiplier=1, max=self.RETRY_MAX_WAIT),
-            before_sleep=self._log_retry_attempt
+            before_sleep=self._log_retry_attempt,
         )
         if retry_period >= 0:
             self.retry_kwargs.update(dict(stop=stop_after_delay(retry_period)))
@@ -82,9 +93,9 @@ class RetryingConnection(object):
     @staticmethod
     def _log_retry_attempt(retry_state):
         """Log when next reconnection attempt is about to be made."""
-        logger.log(logging.WARNING,
-                   "Retrying connection in %s seconds...",
-                   retry_state.next_action.sleep)
+        logger.log(
+            logging.WARNING, "Retrying connection in %s seconds...", retry_state.next_action.sleep
+        )
 
 
 def ensure_conf(app):
@@ -100,6 +111,7 @@ def ensure_conf(app):
     except AttributeError:
         config = RedBeatConfig(app)
         setattr(app, name, config)
+
     return config
 
 
@@ -107,17 +119,31 @@ def get_redis(app=None):
     app = app_or_default(app)
     conf = ensure_conf(app)
     if not hasattr(app, 'redbeat_redis') or app.redbeat_redis is None:
-        redis_options = conf.app.conf.get(
-            'REDBEAT_REDIS_OPTIONS',
-            conf.app.conf.get('BROKER_TRANSPORT_OPTIONS', {}))
+        redis_options = conf.redbeat_redis_options
         retry_period = redis_options.get('retry_period')
-        if conf.redis_url.startswith('redis-sentinel') and  'sentinels' in redis_options:
+        if conf.redis_url.startswith('redis-sentinel') and 'sentinels' in redis_options:
             from redis.sentinel import Sentinel
-            sentinel = Sentinel(redis_options['sentinels'],
-                                socket_timeout=redis_options.get('socket_timeout'),
-                                password=redis_options.get('password'),
-                                decode_responses=True)
+
+            sentinel = Sentinel(
+                redis_options['sentinels'],
+                socket_timeout=redis_options.get('socket_timeout'),
+                password=redis_options.get('password'),
+                db=redis_options.get('db', 0),
+                decode_responses=True,
+                sentinel_kwargs=redis_options.get('sentinel_kwargs'),
+            )
             connection = sentinel.master_for(redis_options.get('service_name', 'master'))
+        elif conf.redis_url.startswith('rediss'):
+            ssl_options = {'ssl_cert_reqs': ssl.CERT_REQUIRED}
+            if isinstance(conf.redis_use_ssl, dict):
+                ssl_options.update(conf.redis_use_ssl)
+            connection = StrictRedis.from_url(conf.redis_url, decode_responses=True, **ssl_options)
+        elif conf.redis_url.startswith('redis-cluster'):
+            from rediscluster import RedisCluster
+
+            if not redis_options.get('startup_nodes'):
+                redis_options = {'startup_nodes': [{"host": "localhost", "port": "30001"}]}
+            connection = RedisCluster(decode_responses=True, **redis_options)
         else:
             connection = StrictRedis.from_url(conf.redis_url, decode_responses=True)
 
@@ -129,14 +155,6 @@ def get_redis(app=None):
     return app.redbeat_redis
 
 
-if REDIS_3_OR_GREATER:
-    def zadd(pipe, name, mapping):
-        pipe.zadd(name, mapping)
-else:
-    def zadd(pipe, name, mapping):
-        pipe.zadd(name, **mapping)
-
-
 ADD_ENTRY_ERROR = """\
 
 Couldn't add entry %r to redis schedule: %r. Contents: %r
@@ -154,37 +172,25 @@ class RedBeatConfig(object):
         self.lock_key = self.either_or('redbeat_lock_key', self.key_prefix + ':lock')
         self.lock_timeout = self.either_or('redbeat_lock_timeout', None)
         self.redis_url = self.either_or('redbeat_redis_url', app.conf['BROKER_URL'])
-
-        self.validate_timezone()
-
-    def validate_timezone(self):
-        tz = self.app.timezone
-
-        if tz == pytz.utc:
-            return
-
-        warnings.warn('Celery configured with %s, RedBeat only supports UTC' % (tz))
+        self.redis_use_ssl = self.either_or('redbeat_redis_use_ssl', app.conf['BROKER_USE_SSL'])
+        self.redbeat_redis_options = self.either_or(
+            'redbeat_redis_options', app.conf['BROKER_TRANSPORT_OPTIONS']
+        )
 
     @property
     def schedule(self):
-        if CELERY_4_OR_GREATER:
-            return self.app.conf.beat_schedule
-        else:
-            return self.app.conf.CELERYBEAT_SCHEDULE
+        return self.app.conf.beat_schedule
 
     @schedule.setter
     def schedule(self, value):
-        if CELERY_4_OR_GREATER:
-            self.app.conf.beat_schedule = value
-        else:
-            self.app.conf.CELERYBEAT_SCHEDULE = value
+        self.app.conf.beat_schedule = value
 
     def either_or(self, name, default=None):
-        if CELERY_4_OR_GREATER and name == name.upper():
+        if name == name.upper():
             warnings.warn(
                 'Celery v4 installed, but detected Celery v3 '
                 'configuration %s (use %s instead).' % (name, name.lower()),
-                UserWarning
+                UserWarning,
             )
         return self.app.conf.first(name, name.upper()) or default
 
@@ -192,10 +198,12 @@ class RedBeatConfig(object):
 class RedBeatSchedulerEntry(ScheduleEntry):
     _meta = None
 
-    def __init__(self, name=None, task=None, schedule=None,
-                 args=None, kwargs=None, enabled=True, **clsargs):
-        super(RedBeatSchedulerEntry, self).__init__(name=name, task=task, schedule=schedule,
-                                                    args=args, kwargs=kwargs, **clsargs)
+    def __init__(
+        self, name=None, task=None, schedule=None, args=None, kwargs=None, enabled=True, **clsargs
+    ):
+        super(RedBeatSchedulerEntry, self).__init__(
+            name=name, task=task, schedule=schedule, args=args, kwargs=kwargs, **clsargs
+        )
         self.enabled = enabled
         ensure_conf(self.app)
 
@@ -270,6 +278,7 @@ class RedBeatSchedulerEntry(ScheduleEntry):
 
     @property
     def score(self):
+        """ return UTC based UNIX timestamp """
         if self.due_at is None:
             # Scores < zero are ignored on each tick.
             return -1
@@ -289,9 +298,13 @@ class RedBeatSchedulerEntry(ScheduleEntry):
             'schedule': self.schedule,
             'enabled': self.enabled,
         }
+        meta = {
+            'last_run_at': self.last_run_at,
+        }
         with get_redis(self.app).pipeline() as pipe:
             pipe.hset(self.key, 'definition', json.dumps(definition, cls=RedBeatJSONEncoder))
-            zadd(pipe, self.app.redbeat_conf.schedule_key, {self.key: self.score})
+            pipe.hsetnx(self.key, 'meta', json.dumps(meta, cls=RedBeatJSONEncoder))
+            pipe.zadd(self.app.redbeat_conf.schedule_key, {self.key: self.score})
             pipe.execute()
 
         return self
@@ -316,10 +329,11 @@ class RedBeatSchedulerEntry(ScheduleEntry):
 
         with get_redis(self.app).pipeline() as pipe:
             pipe.hset(self.key, 'meta', json.dumps(meta, cls=RedBeatJSONEncoder))
-            zadd(pipe, self.app.redbeat_conf.schedule_key, {entry.key: entry.score})
+            pipe.zadd(self.app.redbeat_conf.schedule_key, {entry.key: entry.score})
             pipe.execute()
 
         return entry
+
     __next__ = next = _next_instance
 
     def reschedule(self, last_run_at=None):
@@ -329,15 +343,16 @@ class RedBeatSchedulerEntry(ScheduleEntry):
         }
         with get_redis(self.app).pipeline() as pipe:
             pipe.hset(self.key, 'meta', json.dumps(meta, cls=RedBeatJSONEncoder))
-            zadd(pipe, self.app.redbeat_conf.schedule_key, {self.key: self.score})
+            pipe.zadd(self.app.redbeat_conf.schedule_key, {self.key: self.score})
             pipe.execute()
 
     def is_due(self):
         if not self.enabled:
             return False, 5.0  # 5 second delay for re-enable.
 
-        return self.schedule.is_due(self.last_run_at or
-                                    datetime(MINYEAR, 1, 1, tzinfo=self.schedule.tz))
+        return self.schedule.is_due(
+            self.last_run_at or datetime(MINYEAR, 1, 2, tzinfo=self.schedule.tz)
+        )
 
 
 class RedBeatScheduler(Scheduler):
@@ -353,10 +368,12 @@ class RedBeatScheduler(Scheduler):
     def __init__(self, app, lock_key=None, lock_timeout=None, **kwargs):
         ensure_conf(app)  # set app.redbeat_conf
         self.lock_key = lock_key or app.redbeat_conf.lock_key
-        self.lock_timeout = (lock_timeout or
-                             app.redbeat_conf.lock_timeout or
-                             self.max_interval * 5 or
-                             self.lock_timeout)
+        self.lock_timeout = (
+            lock_timeout
+            or app.redbeat_conf.lock_timeout
+            or self.max_interval * 5
+            or self.lock_timeout
+        )
         super(RedBeatScheduler, self).__init__(app, **kwargs)
 
     def setup_schedule(self):
@@ -379,8 +396,7 @@ class RedBeatScheduler(Scheduler):
 
             # keep track of static schedule entries,
             # so we notice when any are removed at next startup
-            client.sadd(self.app.redbeat_conf.statics_key,
-                        *self.app.redbeat_conf.schedule.keys())
+            client.sadd(self.app.redbeat_conf.statics_key, *self.app.redbeat_conf.schedule.keys())
 
     def update_from_dict(self, dict_):
         for name, entry in dict_.items():
@@ -408,13 +424,16 @@ class RedBeatScheduler(Scheduler):
             pipe.zrangebyscore(self.app.redbeat_conf.schedule_key, 0, max_due_at)
 
             # peek into the next tick to accuratly calculate sleep between ticks
-            pipe.zrangebyscore(self.app.redbeat_conf.schedule_key,
-                               '({}'.format(max_due_at),
-                               max_due_at + self.max_interval,
-                               start=0, num=1)
+            pipe.zrangebyscore(
+                self.app.redbeat_conf.schedule_key,
+                '({}'.format(max_due_at),
+                max_due_at + self.max_interval,
+                start=0,
+                num=1,
+            )
             due_tasks, maybe_due = pipe.execute()
 
-        logger.info('Loading %d tasks', len(due_tasks) + len(maybe_due))
+        logger.debug('Loading %d tasks', len(due_tasks) + len(maybe_due))
         d = {}
         for key in due_tasks + maybe_due:
             try:
@@ -444,11 +463,11 @@ class RedBeatScheduler(Scheduler):
     def tick(self, min=min, **kwargs):
         if self.lock:
             logger.debug('beat: Extending lock...')
-            get_redis(self.app).pexpire(self.lock_key, int(self.lock_timeout * 1000))
+            self.lock.extend(int(self.lock_timeout))
 
         remaining_times = []
         try:
-            for entry in values(self.schedule):
+            for entry in self.schedule.values():
                 next_time_to_run = self.maybe_due(entry, **self._maybe_due_kwargs)
                 if next_time_to_run:
                     remaining_times.append(next_time_to_run)
@@ -468,17 +487,17 @@ class RedBeatScheduler(Scheduler):
     def info(self):
         info = ['       . redis -> {}'.format(maybe_sanitize_url(self.app.redbeat_conf.redis_url))]
         if self.lock_key:
-            info.append('       . lock -> `{}` {} ({}s)'.format(
-                self.lock_key, humanize_seconds(self.lock_timeout), self.lock_timeout))
+            info.append(
+                '       . lock -> `{}` {} ({}s)'.format(
+                    self.lock_key, humanize_seconds(self.lock_timeout), self.lock_timeout
+                )
+            )
         return '\n'.join(info)
 
     @cached_property
     def _maybe_due_kwargs(self):
         """ handle rename of publisher to producer """
-        if CELERY_4_OR_GREATER:
-            return {'producer': self.producer}  # celery 4.x
-        else:
-            return {'publisher': self.publisher}  # celery 3.x
+        return {'producer': self.producer}
 
 
 @beat_init.connect
@@ -488,11 +507,16 @@ def acquire_distributed_beat_lock(sender=None, **kwargs):
         return
 
     logger.debug('beat: Acquiring lock...')
+    redis_client = get_redis(scheduler.app)
 
-    lock = get_redis(scheduler.app).lock(
+    lock = redis_client.lock(
         scheduler.lock_key,
         timeout=scheduler.lock_timeout,
         sleep=scheduler.max_interval,
     )
+    # overwrite redis-py's extend script
+    # which will add additional timeout instead of extend to a new timeout
+    lock.lua_extend = redis_client.register_script(LUA_EXTEND_TO_SCRIPT)
     lock.acquire()
+
     scheduler.lock = lock
